# -*-encoding:utf-8-*-


class HttpConnection(object):
    def __init__(self):
        self.connection = {}
        self.requests = {}
        self.responses = {}

    def add_connection(self, name, value):
        self.connection[name] = value

    def get_connection(self, name):
        return self.connection[name]

    def remove_connection(self, name):
        del self.connection[name]



# -*-coding:utf-8-*-

from jinja2 import Template
import sys

reload(sys)
sys.setdefaultencoding('utf-8')

__author__ = 'karlvorndoenitz@gmail.com'


def render(template_path, **kwargs):
    template_file = open(template_path)
    template_data = template_file.read()
    template = Template(template_data)
    data = template.render(kwargs)
    template_file.close()
    return data


def render_string(template_string, **kwargs):
    template = Template(template_string)
    return template.render(kwargs)



# -*-encoding:utf-8-*-


class HttpIOBuffer(object):

    def __init__(self):
        self.requests = {}
        self.responses = {}

    def add_request(self, name, value):
        self.requests[name] = value

    def add_response(self, name, value):
        self.responses[name] = value

    def get_request(self, name):
        return self.requests[name]

    def get_response(self, name):
        return self.responses[name]

    def remove_request(self, name):
        del self.requests[name]

    def remove_response(self, name):
        del self.responses[name]



# -*-coding:utf-8-*-

"""

"""

__author__ = 'karlvorndoenitz@gmail.com'
__version__ = '0.3.8'
__packages__ = [
    "autoreload",
    "config",
    "escape",
    "http_parser",
    "logger",
    "router",
    "template",
    "utils",
    "web"
]



# -*-coding:utf-8-*-

import datetime
import json
import logging
from urllib import unquote, unquote_plus

from karlooper.utils.http_utils import get_http_content_type
from karlooper.config.config import ContentType, COOKIE_SECURITY_DEFAULT_STRING, HttpStatus, HttpStatusMsg
from karlooper.escape import utf8
from karlooper.utils.encrypt import StrEncryption
from karlooper.template import render

__author__ = 'karlvorndoenitz@gmail.com'


class Request(object):
    def __init__(self, http_data_dict, http_message, settings):
        """

        :param http_data_dict: converted http data, dict type
        :param http_message: http message, string type

        """
        self.__http_data = http_data_dict
        self.header = "ServerName: karlooper\r\n"
        self.__http_message = http_message
        self.logger = logging.getLogger()
        self.cookie_dict = self.__parse_cookie()
        self.param_dict = self.__parse_param()
        self.__settings = settings
        self.__response_cookie = {}
        self.__response_header = {}
        self.__path_param = {}

    def __parse_cookie(self):
        """parse cookie string to cookie dict

        :return: a dict contain cookie or None or error

        """
        cookie_string = self.__http_data["header"].get("cookie", "")
        if not cookie_string:
            return {}
        try:
            cookie_dict = dict((cookie.split("=")[0], cookie.split("=", 1)[1]) for cookie in cookie_string.split("; "))
            return cookie_dict
        except Exception, e:
            raise e

    def __parse_param(self):
        """parse http message to get the param dict

        :return: a dict contain params

        """
        url_param = self.__http_data["url"].split("?")[1] if "?" in self.__http_data["url"] else None
        if not url_param:
            url_param_dict = {}
        else:
            url_param_dict = dict((param.split("=")[0], param.split("=")[1]) for param in url_param.split("&"))
        content_type = self.get_header("content-type")
        content_type = get_http_content_type(content_type)
        http_body = self.__http_data.get("body", "")
        if content_type == ContentType.FORM and http_body:
            body_param = dict((param.split("=")[0], param.split("=")[1]) for param in http_body.split("&"))
        elif content_type == ContentType.JSON and http_body:
            try:
                body_param = json.loads(http_body)
            except Exception, e:
                self.logger.warning("parse json error: %s" % str(e))
                body_param = eval(http_body)
        else:
            body_param = {}
        param = dict(url_param_dict, **body_param)
        return param

    def set_path_param(self, path_param_dict):
        """ set the param in url

        :param path_param_dict: param parsed from url
        :return: None

        """
        self.__path_param = path_param_dict

    def get_path_param(self, key, default=None):
        """ get param in url with key

        :param key: param's key
        :param default: param's default value
        :return: param's value

        """
        return self.__path_param.get(key, default)

    def get_path_param_int(self, key, default=None):
        """ get param's int value in url with key

        :param key: param's key
        :param default: param's default value
        :return: param's value

        """
        value = self.__path_param.get(key, default)
        if value is None:
            return None
        return int(value)

    def get_path_param_str(self, key, default=None):
        """ get param's str value in url with key

        :param key: param's key
        :param default: param's default value
        :return: param's value

        """
        value = self.__path_param.get(key, default)
        if value is None:
            return None
        return str(value)

    def get_path_param_boolean(self, key, default=None):
        """ get param's boolean value in url with key

        :param key: param's key
        :param default: param's default value
        :return: param's value

        """
        value = self.__path_param.get(key, default)
        if value is None:
            return None
        return bool(value)

    def get_path_param_float(self, key, default=None):
        """ get param's float value in url with key

        :param key: param's key
        :param default: param's default value
        :return: param's value

        """
        value = self.__path_param.get(key, default)
        if value is None:
            return None
        return float(value)

    def get_cookie(self, key, default=None):
        """get cookie's value with defined key

        :param key: cookie's key
        :param default: cookie's default value
        :return: cookie's value

        """
        return self.cookie_dict.get(key, default)

    def get_security_cookie(self, key, default=None):
        """get security cookie's value with decode

        :param key: cookie's key
        :param default: cookie's default value
        :return: cookie's value

        """
        cookie = self.get_cookie(key)
        if not cookie:
            return default
        des = StrEncryption()
        security_key = self.__settings.get("cookie", COOKIE_SECURITY_DEFAULT_STRING)
        des.input_key(security_key)
        return des.decode(cookie)

    def get_parameter(self, key, default=None):
        """get parameter's value with the given key,
        this method can get value in http body or url,
        content-type support: application/json and application/x-www-form-urlencoded

        :param key: param's key
        :param default: param's default value
        :return: param's value

        """
        return self.param_dict.get(key, default)

    def decode_parameter(self, key, default=None):
        """decode parameter with defined key

        :param key: value's key
        :param default: default value
        :return: decoded parameter

        """
        parameter = self.get_parameter(key)
        if not parameter:
            return default
        return unquote(parameter)

    def decode_parameter_plus(self, key, default=None):
        """decode parameter with defined key, include decode plus symbol to space

        :param key: value's key
        :param default: default value
        :return: decoded parameter

        """
        parameter = self.get_parameter(key)
        if not parameter:
            return default
        return unquote_plus(parameter)

    def get_header(self, key, default=None):
        """get http header's value with defined key

        :param key: header data's key
        :param default: default value
        :return: value

        """
        header_data = self.__http_data["header"]
        return header_data.get(key, default)

    def __parse_cookie_dict_to_string(self):
        for cookie in self.__response_cookie.values():
            if cookie.get("domain"):
                cookie_format = 'Set-Cookie: %(key)s=%(value)s; expires=%(expires)s; Path=%(path)s; Domain=%(domain)s'
            else:
                cookie_format = 'Set-Cookie: %(key)s=%(value)s; expires=%(expires)s; Path=%(path)s'
            cookie_string = cookie_format % cookie
            self.header += "%s\r\n" % cookie_string

    def generate_expire_date(self, expires_days):
        """get expire date

        :param expires_days: expires days, int type
        :return: http expires days, string type

        """
        now_time = datetime.datetime.now()
        expires_time = now_time + datetime.timedelta(days=expires_days)
        expires_time = expires_time.strftime("%a, %d %b %Y %H:%M:%S GMT")
        self.logger.info(expires_time)
        return expires_time

    def get_now_time(self):
        now_time = datetime.datetime.now().strftime("%a, %d %b %Y %H:%M:%S GMT")
        self.logger.info(now_time)
        return now_time

    def get_request_url(self):
        """get request url

        :return: request url

        """
        return self.__http_data.get("url", "")

    def set_cookie(self, key, value, expires_days=1, path="/", domain=None):
        """set cookie's value

        :param key: cookie's key
        :param value: cookie's value
        :param expires_days: cookie's expires days
        :param path: cookie's value path
        :param domain: cookie's domain
        :return: None

        """
        key = str(key)
        value = str(value)
        expires_days = self.generate_expire_date(expires_days)
        cookie_dict = {
            "key": key,
            "value": value,
            "expires": expires_days,
            "path": path,
            "domain": domain
        }
        if domain:
            cookie_dict["Domain"] = domain
        self.__response_cookie[key] = cookie_dict

    def set_security_cookie(self, key, value, expires_days=1, path="/", domain=None):
        """set security cookie's value

        :param key: cookie's key
        :param value: cookie's value
        :param expires_days: cookie's expires days
        :param path: cookie's value path
        :param domain: cookie's domain
        :return: None

        """
        des = StrEncryption()
        security_key = self.__settings.get("cookie", COOKIE_SECURITY_DEFAULT_STRING)
        des.input_key(security_key)
        security_value = des.encode(value)
        self.set_cookie(key, security_value, expires_days, path, domain)

    def clear_cookie(self, key, path="/", domain=None):
        """clear cookie with defined key

        :param key: clear cookie by key
        :param path: cookie's path
        :param domain: cookie's domain
        :return: None

        """
        self.set_cookie(key, "", 0, path, domain)

    def clear_all_cookie(self, path="/", domain=None):
        """clear all cookie

        :param path: cookie's path
        :param domain: cookie's domain
        :return: None

        """
        for key in self.cookie_dict:
            self.clear_cookie(key, path, domain)

    def __parse_header_dict_to_string(self):
        for header_key in self.__response_header:
            self.header += "%s: %s\r\n" % (header_key, self.__response_header[header_key])

    def set_header(self, header_dict):
        """set header data in http message

        :param header_dict: http header data dict type
        :return: None

        """
        for header_key in header_dict:
            self.__response_header[header_key] = header_dict[header_key]

    def clear_header(self, name):
        """clear headers' data in http response message

        :param name: header's name
        :return:

        """
        self.__response_header.pop(name)

    def get_response_header(self):
        """get http response header

        :return: http message's header

        """
        self.__parse_header_dict_to_string()
        self.__parse_cookie_dict_to_string()
        return "\r\n" + self.header + "\r\n"

    def response_as_json(self, data):
        """decorate data to http json data

        :param data: the response data
        :return: json data

        """
        self.set_header({"Content-Type": "application/json"})
        response = json.dumps(data, ensure_ascii=False)
        return utf8(response), HttpStatus.SUCCESS, HttpStatusMsg.SUCCESS

    def http_response(self, data):
        """decorate data to http response data

        :param data: http response data
        :return: a tuple contains http message, status, status message

        """
        self.logger.info("response data:", data)
        return data, HttpStatus.SUCCESS, HttpStatusMsg.SUCCESS

    def render(self, template_path, **kwargs):
        """render a template

        :param template_path: template's relative path
        :param kwargs: params used in template
        :return: a tuple contains http message, status, status message

        """
        root_path = self.__settings.get("template", ".")
        template_path = root_path + template_path
        return render(template_path, **kwargs), HttpStatus.SUCCESS, HttpStatusMsg.SUCCESS

    def redirect(self, url, status=HttpStatus.REDIRECT):
        """redirect to defined url

        :param url: redirect url
        :param status: http status
        :return: a tuple contains http message, status, status message

        """
        self.set_header({
            "Location": url,
        })
        return "", status, HttpStatusMsg.REDIRECT

    def get_http_request_message(self):
        """get request message

        :return: self http message

        """
        return self.__http_message

    def get(self):
        self.logger.error(HttpStatusMsg.METHOD_NOT_ALLOWED)
        return "405", HttpStatus.METHOD_NOT_ALLOWED, HttpStatusMsg.METHOD_NOT_ALLOWED

    def post(self):
        self.logger.error(HttpStatusMsg.METHOD_NOT_ALLOWED)
        return "405", HttpStatus.METHOD_NOT_ALLOWED, HttpStatusMsg.METHOD_NOT_ALLOWED

    def put(self):
        self.logger.error(HttpStatusMsg.METHOD_NOT_ALLOWED)
        return "405", HttpStatus.METHOD_NOT_ALLOWED, HttpStatusMsg.METHOD_NOT_ALLOWED

    def head(self):
        self.logger.info(self.__http_data.get("url", ""))
        return ""

    def options(self):
        self.logger.error(HttpStatusMsg.METHOD_NOT_ALLOWED)
        return "405", HttpStatus.METHOD_NOT_ALLOWED, HttpStatusMsg.METHOD_NOT_ALLOWED

    def delete(self):
        self.logger.error(HttpStatusMsg.METHOD_NOT_ALLOWED)
        return "405", HttpStatus.METHOD_NOT_ALLOWED, HttpStatusMsg.METHOD_NOT_ALLOWED

    def trace(self):
        return self.__http_message.split("\r\n\r\n") if "\r\n\r\n" in self.__http_message else ""

    def connect(self):
        self.logger.error(HttpStatusMsg.METHOD_NOT_ALLOWED)
        return "405", HttpStatus.METHOD_NOT_ALLOWED, HttpStatusMsg.METHOD_NOT_ALLOWED

    def before_request(self):
        pass

    def teardown_request(self):
        pass



# -*-coding:utf-8-*-

import re
import urllib
import htmlentitydefs

_XHTML_ESCAPE_RE = re.compile('[&<>"\']')
_XHTML_ESCAPE_DICT = {'&': '&amp;', '<': '&lt;', '>': '&gt;', '"': '&quot;', '\'': '&#39;'}
_BASESTRING_TYPES = (basestring, type(None))
_TO_UNICODE_TYPES = (type(u''), type(None))
_unicode_type = type(u'')


def to_unicode(value):
    """Converts a string argument to a unicode string.

    :param value: value to unicode
    :return: unicode value

    """
    if isinstance(value, _TO_UNICODE_TYPES):
        return value
    if not isinstance(value, bytes):
        raise TypeError(
            "Expected bytes, unicode, or None; got %r" % type(value)
        )
    return value.decode("utf-8")


def to_basestring(value):
    """Converts a string argument to a subclass of basestring.

    :param value: the value will to be basestring
    :return: basestring value

    """
    if isinstance(value, _BASESTRING_TYPES):
        return value
    if not isinstance(value, bytes):
        raise TypeError(
            "Expected bytes, unicode, or None; got %r" % type(value)
        )
    return value.decode("utf-8")


def utf8(value):
    """Converts a string argument to a byte string.

    :param value: value will encode to utf-8
    :return: utf-8 value

    """
    if isinstance(value, (bytes, type(None))):
        return value
    if not isinstance(value, _TO_UNICODE_TYPES):
        raise TypeError(
            "Expected bytes, unicode, or None; got %r" % type(value)
        )
    return value.encode("utf-8")


def _build_unicode_map():
    unicode_map = {}
    for name, value in htmlentitydefs.name2codepoint.items():
        unicode_map[name] = unichr(value)
    return unicode_map


def _convert_entity(m):
    if m.group(1) == "#":
        try:
            if m.group(2)[:1].lower() == 'x':
                return unichr(int(m.group(2)[1:], 16))
            else:
                return unichr(int(m.group(2)))
        except ValueError:
            return "&#%s;" % m.group(2)
    try:
        return _build_unicode_map()[m.group(2)]
    except KeyError:
        return "&%s;" % m.group(2)


def xhtml_escape(value):
    """Escapes a string so it is valid within HTML or XML.

    :param value: the value will be escaped
    :return: escaped value

    """
    return _XHTML_ESCAPE_RE.sub(
        lambda match: _XHTML_ESCAPE_DICT[match.group(0)],
        to_basestring(value)
    )


def xhtml_unescape(value):
    """Un-escapes an XML-escaped string.

    :param value: value will be escaped
    :return: unescaped value

    """
    return re.sub(r"&(#?)(\w+?);", _convert_entity, to_unicode(value))


def url_escape(value, plus=True):
    """Returns a URL-encoded version of the given value.

    :param value: url will be escaped
    :param plus: default True, whether need plus
    :return: the escaped value

    """
    quote = urllib.quote_plus if plus else urllib.quote
    return quote(utf8(value))


def url_unescape(value, encoding='utf-8', plus=True):
    """Decodes the given value from a URL.

    :param value: value will be unescaped
    :param encoding: the encoding default is utf-8
    :param plus: whether need plus default True
    :return: the unescaped url

    """
    unquote = (urllib.unquote_plus if plus else urllib.unquote)
    if encoding is None:
        return unquote(utf8(value))
    else:
        return _unicode_type(unquote(utf8(value)), encoding)



# -*-coding:utf-8-*-
from karlooper.config import get_cli_data
from karlooper.web.statics import StaticHandler


class PathParam(object):
    status = False
    value = {}


class Router(object):
    def __init__(self, handles, url):
        self.__handlers = handles
        self.__url = url
        self.__is_static_enable = get_cli_data().get("static")
        self.__is_debug = get_cli_data().get("debug")

    def get_handler(self):
        """
        I will add url regex parse in the future
        :return: return a handler object
        """
        path_param = PathParam()
        handler = self.__handlers.get(self.__url, None)
        if handler is None:
            url_list = self.__handlers.keys()
            for url in url_list:
                path_param = self.get_path_param(url)
                if path_param.status:
                    handler = self.__handlers.get(url, None)
                    return handler, path_param
        if handler is None \
                and self.__is_debug \
                and self.__is_static_enable \
                and (str(self.__is_debug).lower() == "true") \
                and (not handler) \
                and ("/static/" in self.__url):
            handler = StaticHandler
        return handler, path_param

    def get_path_param(self, path):
        """ parse param in url

        :param path: url rule
        :return: a PathParam object

        """

        path_attr_list = path.split("/")
        url_attr_list = self.__url.split("/")
        path_param = PathParam()
        if len(path_attr_list) != len(url_attr_list):
            path_param.status = False
            return path_param
        for index in xrange(len(path_attr_list)):
            path_attr = path_attr_list[index]
            url_attr = url_attr_list[index]
            if ("{" not in path_attr or "}" not in path_attr) and path_attr != url_attr:
                path_param.status = False
                return path_param
            if "{" in path_attr and "}" in path_attr:
                path_value_key = path_attr.replace("{", "").replace("}", "")
                path_param.status = True
                path_param.value[path_value_key] = url_attr
        return path_param



# -*-coding:utf-8-*-
"""

statics
~~~~~~~

introduction
a simple static http handler

Warning: this model just support debug model, we do not recommend developer use this models,
we recommend developer to use nginx as a static server.


"""

import os
import hashlib
from karlooper.web.request import Request
from karlooper.config import get_cli_data
from karlooper.config.config import content_type, HttpStatus, HttpStatusMsg


class StaticHandler(Request):
    """

    Two methods:

    def get(self): http get method

    def get_file_etag(self, file_path): get file's etag

    """
    def get(self):
        """

        :return: http response data, status, status message

        """
        global_config_data = get_cli_data()
        static_root = global_config_data.get("static", ".")
        request_expire_days = self.get_header("expires")
        now_time = self.get_now_time()
        if request_expire_days and now_time <= request_expire_days:
            return "", HttpStatus.RESOURCE_NOT_MODIFIED, HttpStatusMsg.RESOURCE_NOT_MODIFIED
        request_etag = self.get_header("if-none-match")
        file_path = self.get_request_url()
        file_absolute_path = static_root+file_path
        if not os.path.exists(file_absolute_path):
            return "404", HttpStatus.NOT_FOUND, HttpStatusMsg.NOT_FOUND
        file_etag, file_data = self.get_file_etag(file_absolute_path)
        if request_etag and request_etag == file_etag:
            return "", HttpStatus.RESOURCE_NOT_MODIFIED, HttpStatusMsg.RESOURCE_NOT_MODIFIED
        expires = self.generate_expire_date(expires_days=7)
        file_extension = file_path.split(".")[-1]
        self.set_header({
            "ETag": file_etag,
            "Expires": expires,
            "Content-Type": content_type.get(file_extension)
        })
        return file_data, HttpStatus.SUCCESS, HttpStatusMsg.SUCCESS

    def get_file_etag(self, file_path):
        """get file's etag

        :param file_path: static file's path
        :return: file's etag

        """
        f = open(file_path)
        file_data = f.read()
        f.close()
        etag = hashlib.md5(file_data).hexdigest()
        self.logger.info("%s's etag is %s" % (file_path, etag))
        return etag, file_data



#!/usr/bin/env python
from __future__ import print_function
from setuptools import setup
from karlooper import __version__

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='karlooper',
    version=__version__,
    install_requires=required,
    packages=[
        'karlooper',
        'karlooper.autoreload',
        'karlooper.config',
        'karlooper.escape',
        'karlooper.http_parser',
        'karlooper.logger',
        'karlooper.router',
        'karlooper.template',
        'karlooper.utils',
        'karlooper.web'
    ],
    data_files=[
        ('lib/python2.7/dist-packages/karlooper/logger', ['karlooper/logger/log.conf']),
        ('lib/python2.7/site-packages/karlooper/logger', ['karlooper/logger/log.conf'])
    ],
    url='https://github.com/karldoenitz/karlooper',
    license='',
    author='lizhihao',
    author_email='karlvorndoenitz@gmail.com',
    description='karloop plus framework',
    long_description=open("README.rst").read(),
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7'
    ]
)



# -*-coding:utf-8-*-
"""

parse_command_line
~~~~~~~~~~~~~~~~~~

introduction
first initial an object of CommandLineParser
then use the object to set some default values
end, use parse_command_line function to parse the command line

Usage
=====
>>> from karlooper.utils.parse_command_line import CommandLineParser
>>> CommandLineParser.default(port=8080, debug=True, log_enable=False)
>>> CommandLineParser.parse_command_line()

"""

import sys
from karlooper.config import set_cli_data


message = """
Usage: python index.py [name]=[value]
           (to execute a python file)
where options include:
    help: for help
    port: set port that the server to listen
    is_debug: True or False, set True to enable debug
    log_enable: True or False, set True to enable log
    cookie: cookie's security key
"""


class CommandLineParser(object):
    """

    Two methods:

    def default(**kwargs):  set default arguments

    def parse_command_line():  parse the command line to get arguments

    """
    @staticmethod
    def default(**kwargs):
        set_cli_data(kwargs)

    @staticmethod
    def parse_command_line():
        params = sys.argv[1:]
        __params_dict = dict({})
        if "help" in params:
            print message
            sys.exit(0)
        for param in params:
            __params_dict[param.split("=")[0]] = param.split("=")[1]
        set_cli_data(data=__params_dict)
